// Library for reading and writing of structured strings
// Standard created by: Medwyn_cz, Gord
// Coded by: Medwyn_cz
// Version: 1.1

// Format
// structString: <name1="value1"/><name2="value2"/><name3="value3"/>...
// structStringList: name1,value1,name2,value2,name3,value3...

// Example
// structString: <owner="medwyn_cz"/><owner="gord"/><project="StructString"/>
// structStringList: owner,medwyn_cz,owner,gord,project,StructString

// all names and values always contain " encoded as &quote&
// all CRLF values are encoded into &CrLf&

const
  REGEX_NAME_VALUE_PAIR = '<\s*.+?\s*=\s*".+?"\s*/\s*>';
  REGEX_NAME_ROUGH = '<\s*.+?\s*=';
  REGEX_VALUE_ROUGH = '=\s*".+?"\s*/\s*>';
  
  CRLF_ENCODED = '&CrLf&';
  QUOTE = '"';
  QUOTE_ENCODED = '&quote&';

  
//------------------------------------------------------------------------------
//FUNCTIONS FOR ENCODING AND DECODING OF STRING INTO/FROM COMPATIBLE FORMAT ----
//------------------------------------------------------------------------------

{Function encodes string into structString code}
function StructStringEncode(str : String) : String;
begin
    Result := str;
    Result := ReplaceString(Result,CRLF,CRLF_ENCODED);
    Result := ReplaceString(Result,QUOTE,QUOTE_ENCODED);
end;
    
{Function decodes string from structString code}
function StructStringDecode(str : String) : String;
begin
    Result := str;
    Result := ReplaceString(Result,CRLF_ENCODED,CRLF);
    Result := ReplaceString(Result,QUOTE_ENCODED,QUOTE);
end;

//------------------------------------------------------------------------------
//FUNCTIONS FOR READING AND WRITING OF THE AGGREGATED STRING/LIST --------------
//------------------------------------------------------------------------------

{Function returns StringList of values name1,value1,name2,value2,name3,value3 ...}
function StructStringGetList(structString: String): TStrings;
var pair,pairs,name,value : String;
begin
    Result := TStringList.Create;
    pairs := RegexExtract(REGEX_NAME_VALUE_PAIR,structString); 
    while (Length(pairs)>0) do
      begin 
        pair := fetch(pairs,CRLF);

        name := RegexExtract(REGEX_NAME_ROUGH,pair);
        name := RegexReplace('^<\s*',name,'',false);
        name := RegexReplace('\s*=$',name,'',false);
        name := StructStringEncode(name); 

        value := RegexExtract(REGEX_VALUE_ROUGH,pair);
        value := RegexReplace('=\s*"',value,'',false);
        value := RegexReplace('"\s*/\s*>',value,'',false);
        value := StructStringEncode(value);

        Result.add(name); 
        Result.add(value); 
      end;
end;

{Function returns structString of values from structStringList}
function StructStringGetString(structStringList: TStrings) : String;
var n : Integer;
begin
   Result := '';
   for n := 0 to structStringList.count-1 do
   begin
      if (n mod 2 <> 0) then continue; //only names
      Result := Result + '<' + StructStringEncode(structStringList[n]) + '="' + StructStringEncode(structStringList[n+1])+'"/>'; 
   end;
end;

//------------------------------------------------------------------------------
//FUNCTIONS - GET VALUES -------------------------------------------------------
//------------------------------------------------------------------------------

{Function returns StringList of all values of given name in given structStringList. Values must be decoded using StructStringDecode before use}
function StructStringListGetMultipleValues(name : String;structStringList : TStrings): TStrings;
var n : Integer;
begin
   name := StructStringEncode(name);
   name := Trim(name);
   Result := TStringList.Create
   for n := 0 to structStringList.Count-1 do
   begin
      if (n mod 2 <> 0) then continue; //only names
      if (structStringList[n] = name) then Result.Add(structStringList[n+1]);
   end;
end;

{Function returns StringList of all values of given name in given structString. Values must be decoded using StructStringDecode before use}
function StructStringGetMultipleValues(name,structString : String): TStrings;
begin
  Result := StructStringListGetMultipleValues(name,StructStringGetList(structString));
end;

{Function returns value of first occurence of value of given name in given structStringList}
function StructStringListGetValue(name : String;structStringList : TStrings) : String;
var list : TStrings;
begin
   list := StructStringListGetMultipleValues(name,structStringList);
   try
   if (list.count = 0) 
     then Result := ''
     else Result := list[0];
     Result := StructStringDecode(Result);
   finally
     list.Free;
   end; 
end;

{Function returns value of first occurence of value of given name in given structString}
function StructStringGetValue(name,structString : String) : String;
begin
  Result := StructStringListGetValue(name,StructStringGetList(structString));
end;

//------------------------------------------------------------------------------
//FUNCTIONS - DELETE VALUES ----------------------------------------------------
//------------------------------------------------------------------------------

{Function deletes all values of given name and value in given structStringList}
procedure StructStringListDeleteValue(name, value : String; var structStringList : TStrings);
var n : Integer;
begin
   name := StructStringEncode(name);
   name := Trim(name);
   value := StructStringEncode(value);
   for n := structStringList.Count-1 downto 0 do
   begin
      if (n mod 2 <> 0) then continue; //only names
      if ((structStringList[n] = name) and ((structStringList[n+1] = value) or (value = '*'))) then
      begin
          structStringList.delete(n+1);
          structStringList.delete(n);
      end;
   end;
end;

{Function deletes all values of given name and value in given structString}
procedure StructStringDeleteValue(name, value : String; var structString : String);
var l : TStrings;
begin
   l := TStringList.Create;
   try
      l := StructStringGetList(structString);
      StructStringListDeleteValue(name, value,l);
      structString := StructStringGetString(l);
   finally
      l.Free;
   end;
end; 

{Function deletes all values of given name in given structStringList}
procedure StructStringListDeleteMultipleValues(name : String; var structStringList :TStrings);
begin
   StructStringListDeleteValue(name,'*', structStringList); 
end;

{Function deletes all values of given name in given structString}
procedure StructStringDeleteMultipleValues(name : String; var structString : String);
var l : TStrings;
begin
   l := TStringList.Create;
   try
      l := StructStringGetList(structString);
      StructStringListDeleteMultipleValues(name,l);
      structString := StructStringGetString(l);
   finally
      l.Free;
   end;
end;

//------------------------------------------------------------------------------
//FUNCTIONS - SET VALUES -------------------------------------------------------
//------------------------------------------------------------------------------

{Function sets values of given name in the given structStringList. If overwrite is true, all other values of given name will be deleted before}
procedure StructStringListSetMultipleValues(name : String; value : TStrings ; var structStringList :TStrings; overwrite: boolean);
var n : Integer;
begin
    name := StructStringEncode(name);
    name := Trim(name);
    for n := 0 to value.count-1 do value[n] := StructStringEncode(value[n]);
    
    if overwrite then StructStringListDeleteMultipleValues(name,structStringList);
    
    for n :=0 to value.count-1 do
    begin
       structStringList.Add(name);
       structStringList.Add(value[n]);
    end;
end;

{Function sets values of given name in the given structString. If overwrite is true, all other values of given name will be deleted before}
procedure StructStringSetMultipleValues(name : String; value :TStrings; var structString: String; overwrite: boolean);
var l : TStrings;
begin
   l := TStringList.Create;
   try
      l := StructStringGetList(structString);
      StructStringListSetMultipleValues(name,value,l,overwrite);
      structString := StructStringGetString(l);
   finally
      l.Free;
   end;
end;

{Function sets value of given name in the given structStringList. If overwrite is true, all other values of given name will be deleted before}
procedure StructStringListSetValue(name,value : String; var structStringList : TStrings; overwrite : boolean);
var l : TStrings;
begin
   l := TStringList.Create;
   try
      l.Add(StructStringEncode(value))
      StructStringListSetMultipleValues(name, l, structStringList, overwrite);
   finally
      l.Free;
   end;
end;

{Function sets value of given name in the given structString. If overwrite is true, all other values of given name will be deleted before}
procedure StructStringSetValue(name,value : String; var structString : String; overwrite: boolean);
var l : TStrings;
begin
   l := TStringList.Create;
   try
      l := StructStringGetList(structString);
      StructStringListSetValue(name,value,l,overwrite);
      structString := StructStringGetString(l);
   finally
      l.Free;
   end;
end;

//------------------------------------------------------------------------------
//FUNCTIONS - QUERY VALUES -----------------------------------------------------
//------------------------------------------------------------------------------

{Does the structStringList contain the given value? If regex is true, value parameter is handled as regular expression}
function StructStringListContains(name,value : String; structStringList : TStrings; regex : boolean) : boolean;
var n : Integer;
begin
    name := StructStringEncode(name);
    name := Trim(name);
    if not regex then value := StructStringEncode(value);
    for n := 0 to structStringList.Count-1 do
    begin
        if (n mod 2 <> 0) then continue; //only names
        if (structStringList[n] = name) then
        begin
            if ((regex and RegexFind(value,structStringList[n+1])) or (structStringList[n+1] = value)) then 
            begin
                Result := true;
                Break;
            end;
        end;
    end;
end;

{Does the structString contain the given value? If regex is true, value parameter is handled as regular expression}
function StructStringContains(name,value,structString : String; regex : boolean) : boolean;
begin
    Result := StructStringListContains(name,value,StructStringGetList(structString),regex);
end;

{Does the structStringList contain value of the given name?}
function StructStringListContainsName(name: String; structStringList : TStrings) : boolean;
begin
    Result := StructStringListContains(name,'(?s).*',structStringList, true);
end;

{Does the structString contain value of the given name?}
function StructStringContainsName(name,structString: String) : boolean;
begin
    Result := StructStringListContainsName(name,StructStringGetList(structString));
end;